/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *             Copyright (C) 2000-2002 by the OpenSG Forum                   *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

//---------------------------------------------------------------------------
//  Includes
//---------------------------------------------------------------------------
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include <set>
#include <sstream>

#include "OSGConfig.h"
#include "OSGLog.h"
#include "OSGBaseThread.h"
#include "OSGSocketSelection.h"
#include "OSGBinaryMessage.h"
#include "OSGGroupMCastConnection.h"
#include "OSGConnectionType.h"
#include "OSGCRC32.h"

#include "OSGminiz.h"
#include "OSGLZ4.h"

//#define MULTI_THREADED

OSG_USING_NAMESPACE

/** \class osg::GroupMCastConnection
 *  \ingroup GrpBaseNetwork
 **/

/*-------------------------------------------------------------------------*/
/*                            constructor destructor                       */

/*! Constructor
 */

GroupMCastConnection::GroupMCastConnection():
    Inherited(),
    _mcastSocket(),
    _mcastAddress(),
    _initialized(false),
    _write_id(0),
    _dgramPool(),
    _write_compression(false)
{
    // prepare mcast socket
    _mcastSocket.open();
    _mcastSocket.bind();
    _mcastSocket.setReadBufferSize(262144);
}

/*! Destructor
 */
GroupMCastConnection::~GroupMCastConnection(void)
{
    // close socket
    _mcastSocket.close();
}

/*! get connection type
 */
const ConnectionType *GroupMCastConnection::getType() const
{
    return &_type;
}

/*-------------------------------------------------------------------------*/
/*                            connection                                   */

/*! connect to the given point. If timeout is reached, -1 is
    returned
 */
GroupConnection::Channel GroupMCastConnection::connectPoint(
    const std::string &address,
    Time               timeout)
{
    Channel channel = Inherited::connectPoint(address,timeout);
    return channel;
}

/*! disconnect the given channel
 */
void GroupMCastConnection::disconnect(Channel channel)
{
    //printf("GroupMCastConnection::disconnect: channel %d\n", channel);
    Inherited::disconnect(channel);
}

/*! accept an icomming point connection. If timeout is reached,
    -1 is returned. If timeout is -1 then wait without timeout
 */
GroupConnection::Channel GroupMCastConnection::acceptPoint(Time timeout)
{
    Connection::Channel channel = Inherited::acceptPoint(timeout);
    return channel;
}

void GroupMCastConnection::flush(void)
{
    GroupSockConnection::flush();
#ifdef MULTI_THREADED
    _tp.wait();
#endif
}

/*! parse the params string.
 */
void GroupMCastConnection::setParams(const std::string &params)
{
    if(params.empty())
        return;

    Inherited::setParams(params);

    std::string option = "TTL=";
    std::string::size_type i = 0;
    if((i=params.find(option)) != std::string::npos)
    {
        std::string str = params.substr(i + option.size());

        std::stringstream ss;
        std::string::size_type j = 0;
        while(j < str.length() && str[j] != ',' && isdigit(str[j]))
        {
            ss << str[j++];
        }
        UInt32 ttl;
        ss >> ttl;
        if(ttl > 255)
            ttl = 255;
        _mcastSocket.setTTL(static_cast<unsigned char>(ttl));
        FINFO(("GroupMCastConnection::setParams : setting ttl to %u.\n", ttl));
    }
}

/*-------------------------------------------------------------------------*/
/*                            sync                                         */

/*! wait for signal
 */
bool GroupMCastConnection::wait(Time timeout) throw (ReadError)
{
    // todo
    return Inherited::wait(timeout);
}

/*! send signal
 */
void GroupMCastConnection::signal(void) throw (WriteError)
{
    UInt32 tag=314156;
    putValue(tag);
    flush();
}

/*-------------------------- helpers --------------------------------------*/

/** \brief create conneciton
 */

GroupConnection *GroupMCastConnection::create(void)
{
    return new GroupMCastConnection();
}

/*-------------------------------------------------------------------------*/
/*                              read write                                 */

/** Write data to all destinations
 *
 * \param mem   Pointer to data buffer
 * \param size  Size of bytes to write
 *
 **/

void GroupMCastConnection::write(const MemoryHandle mem, SizeT size)
{
    if(!_initialized)
        initialize();

    if(size == 0)
        return;

    //printf("write %u\n", size);

    /*
    CRC32 crc;
    UInt32 checksum = crc.fullCRC(mem, size);
    */

    //Dgram  *dgram = new Dgram;
    Dgram *dgram = _dgramPool.alloc();
    UInt32 count = size / dgram->getCapacity();
    count += (dgram->getCapacity() * count) == size ? 0 : 1;

    //printf("send dgrams count %u\n", count);
    std::vector<Int32> remove;
    UInt32 count_id = (_write_id << 16) | (count & 0xffff);
    for(Int32 index = 0 ; index < _sockets.size() ; ++index)
    {
        try
        {
            _sockets[index].send(&count_id, sizeof(count_id));
        }
        catch(SocketException &e)
        {
            remove.push_back(indexToChannel(index));
        }
    }

    if(!remove.empty())
    {
        for(size_t i=0;i<remove.size();++i)
        {
            disconnect(remove[i]);
        }
        if(getChannelCount() == 0)
        {
            throw WriteError("Last running cluster server crashed!");
        }
    }

    //UInt32 ack = 0;
    //for(Int32 index = 0 ; index < _sockets.size() ; ++index)
    //    _sockets[index].recv(&ack, sizeof(ack));

    //printf("send %d dgrams\n", count);

    UInt32  pos    = 0;
    char   *buffer = reinterpret_cast<char*>(mem);
    std::vector<Dgram *> dgrams;
    for(UInt32 i=0;i<count;++i)
    {
        if(dgram == NULL)
        {
            //dgram = new Dgram;
            dgram = _dgramPool.alloc();
        }

        dgram->setBaseId(_write_id);
        dgram->setId(i);
        dgram->setSize(osgMin(UInt32(size), dgram->getCapacity()));
        memcpy(dgram->getData(), buffer + pos, dgram->getSize());

        size -= dgram->getSize();
        pos  += dgram->getSize();

        // with this trick we transfer the complete buffer important for the compression!
        UInt16 flags = (_write_compression ? (Dgram::DGRAM_LAST | Dgram::DGRAM_COMPRESSED) : Dgram::DGRAM_LAST);
        dgram->setFlags(size == 0 ? flags : 0);

        /*
        _mcastSocket.sendTo(dgram->getBuffer(),
                            dgram->getBufferSize(),
                            _mcastAddress);
        */
        // reduce missing dgrams!
        //osgsleep(i % 2);

        //printf("send dgram %u size %u\n", dgram->getId(), dgram->getSize());
        dgrams.push_back(dgram);
        dgram = NULL;
    }

    if(count > _interactive_threshold)
    {
        // we send them twice improves the speed form 2 MB/s to 7 MB/s
        for(UInt32 j=0;j<2;++j)
        {
            for(UInt32 i=0;i<count;++i)
            {
                _mcastSocket.sendTo(dgrams[i]->getBuffer(),
                                    dgrams[i]->getBufferSize(),
                                    _mcastAddress);
            }
        }

        remove.clear();
        UInt32 ack_count = 0;
        for(Int32 index = 0 ; index < _sockets.size() ; ++index)
        {
            try
            {
                _sockets[index].recv(&ack_count, sizeof(ack_count));

                UInt32 missing = count - ack_count;
                //if(ack_count < count)
                //    printf("got only %u dgrams expected %u!\n", ack_count, count);

                while(ack_count < count)
                {
                    UInt32 id = 0;
                    _sockets[index].recv(&id, sizeof(id));

                    //printf("re-send dgram %d\n", id);
                    dgram = dgrams[id];
                    UInt32 dgram_size = dgram->getBufferSize();
                    _sockets[index].send(&dgram_size, sizeof(dgram_size));
                    _sockets[index].send(dgram->getBuffer(), dgram_size);
                    //UInt32 ack = 0;
                    //_sockets[index].recv(&ack, sizeof(ack));
                    ++ack_count;
                    //printf("got re-send ack\n");
                }

                 //_sockets[index].send(&checksum, sizeof(checksum));
                if(missing > 0)
                    printf("channel %u send missing dgrams %u/%u\n", index, missing, count);
            }
            catch(SocketException &e)
            {
                remove.push_back(indexToChannel(index));
            }
        }

        if(!remove.empty())
        {
            for(size_t i=0;i<remove.size();++i)
            {
                disconnect(remove[i]);
            }
            if(getChannelCount() == 0)
            {
                throw WriteError("Last running cluster server crashed!");
            }
        }
    }
    else
    {
        remove.clear();
        for(Int32 index = 0 ; index < _sockets.size() ; ++index)
        {
            try
            {
                for(UInt32 i=0;i<count;++i)
                {
                    dgram = dgrams[i];
                    UInt32 dgram_size = dgram->getBufferSize();
                    _sockets[index].send(&dgram_size, sizeof(dgram_size));
                    _sockets[index].send(dgram->getBuffer(), dgram_size);
                }
            }
            catch(SocketException &e)
            {
                remove.push_back(indexToChannel(index));
            }
        }

        if(!remove.empty())
        {
            for(size_t i=0;i<remove.size();++i)
            {
                disconnect(remove[i]);
            }
            if(getChannelCount() == 0)
            {
                throw WriteError("Last running cluster server crashed!");
            }
        }
    }

    for(size_t i=0;i<dgrams.size();++i)
    {
        //delete dgrams[i];
        _dgramPool.free(dgrams[i]);
    }

    ++_write_id;
}

typedef struct
{
    std::vector<UInt8> buffer;
    UInt32 compression;
    GroupMCastConnection *the;
} ThreadData;

/** Write buffer
 *
 * Write blocksize and data.
 *
 **/
void GroupMCastConnection::writeBuffer(void)
{
    UInt32       size   = writeBufBegin()->getDataSize();
    MemoryHandle buffer = writeBufBegin()->getMem();

    Real64 t = OSG::getSystemTime();

    // we calculate the uncompressed size!
    _write_buffer_bytes += size;

    bool compression = false;

    if(_compression == Connection::CCOMPRESSION_BEST && size > 1)
    {
        static std::vector<UInt8> compressed;
        compressed.resize(mz_compressBound(size));
        mz_ulong compressed_size = compressed.size();
        int r = mz_compress2((unsigned char *) &compressed[0], &compressed_size, (const unsigned char *) buffer, size, 1);
        if(r == 0 && compressed_size < size)
        {
            memcpy(buffer, &compressed[0], compressed_size);
            size = compressed_size;
            compression = true;
        }
    }
    else if(_compression == Connection::CCOMPRESSION_FAST && size > 1)
    {
        static std::vector<UInt8> compressed;
        compressed.resize(LZ4_compressBound(size));
        int compressed_size = LZ4_compress((const char *) buffer, (char *) &compressed[0], size);
        if(compressed_size < size)
        {
            memcpy(buffer, &compressed[0], compressed_size);
            size = compressed_size;
            compression = true;
        }
    }

#ifdef MULTI_THREADED
    _tp.wait();

    static ThreadData *td = new ThreadData;
    td->compression = compression;
    td->buffer.resize(size);
    memcpy(&td->buffer[0], buffer, size);
    td->the = this;
    _tp.run(writeThread, static_cast<void *>(td));
#else
    _write_compression = compression;
    write(buffer, size);
    _write_compression = false;
#endif

    _write_buffer_time += (OSG::getSystemTime() - t);
    _write_buffer_cbytes += size;
}

/*-------------------------------------------------------------------------*/
/*                              private helpers                            */

/*! write thread
 */
void GroupMCastConnection::writeThread(void *arg)
{
    ThreadData *td = (ThreadData *) arg;
    td->the->_write_compression = td->compression;
    td->the->write(&td->buffer[0], td->buffer.size());
    td->the->_write_compression = false;
    // we use one thread only so one td instance is enough!
    //delete td;
}

/*! initialize connection. Connect to all points
 */
void GroupMCastConnection::initialize()
{
    std::string   group = "239.33.42.32";
//    std::string group = "146.140.32.7";
//    std::string group = "146.140.32.255";
    int           port = 15356;
    int           pos  = _destination.find(':');
    int           clientPort;
    std::string   clientHost;
    UInt32        index;
    UInt32        len;
    UInt32        ackNum = UInt32(osgsqrt(Real32(_sockets.size())));
    UInt32        numSource;
    UInt32        sendTo;
    BinaryMessage message;

    if(!getDestination().empty())
        group = getDestination();

    if(_sockets.size()<=16)
        ackNum = 1;

    if(pos>=0)
    {
        group = std::string(_destination,0,pos);
        port  = atoi(std::string(_destination,pos+1,std::string::npos).c_str());
    }
    else
    {
        if(_destination.size())
            group = _destination;
    }
    _mcastAddress.setHost(group);
    _mcastAddress.setPort(port );

    // set multicast interface, if given
    if(!getInterface().empty())
        _mcastSocket.setMCastInterface(
            SocketAddress(getInterface().c_str()));

    for(index = 0 ; index < _sockets.size() ; ++index)
    {
        message.clear();
        // tell the point connection the multicast address
        message.putString(_mcastAddress.getHost());
        message.putUInt32(_mcastAddress.getPort());
        // tell the current seq number
        //message.putUInt32(_seqNumber);
        // tell the point from wich port requests are comming
        message.putUInt32(_mcastSocket.getAddress().getPort());
        // send the message
        _sockets[index].send(message);
    }

    _initialized = true;
}

/*-------------------------------------------------------------------------*/
/*                              static type                                */

ConnectionType GroupMCastConnection::_type(
    &GroupMCastConnection::create,
    "Multicast");

/*-------------------------------------------------------------------------*/
/*                              cvs id's                                   */

#ifdef __sgi
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning( disable : 177 )
#endif

namespace
{
    static Char8 cvsid_cpp       [] = "@(#)$Id: $";
    static Char8 cvsid_hpp       [] = OSG_GROUPMCASTCONNECTION_HEADER_CVSID;
}

