/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *             Copyright (C) 2000-2002 by the OpenSG Forum                   *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

//---------------------------------------------------------------------------
//  Includes
//---------------------------------------------------------------------------
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "OSGConfig.h"
#include "OSGLog.h"
#include "OSGBaseThread.h"
#include "OSGSocketSelection.h"
#include "OSGBinaryMessage.h"
#include "OSGPointMCastConnection.h"
#include "OSGGroupMCastConnection.h"
#include "OSGConnectionType.h"
#include "OSGCRC32.h"

#include "OSGminiz.h"
#include "OSGLZ4.h"

#define AMZ_MCAST

OSG_USING_NAMESPACE

/** \class osg::PointMCastConnection
 *  \ingroup GrpBaseNetwork
 **/

/*-------------------------------------------------------------------------*/
/*                            constructor destructor                       */

/*! Constructor
 */

PointMCastConnection::PointMCastConnection():
    Inherited(),
    _mcastSocket(),
    _sender(),
    _initialized(false),
    _dgramPool()
{
    _acceptSocket.open();
    _acceptSocket.setReusePort(true);
}

/*! Destructor
 */
PointMCastConnection::~PointMCastConnection(void)
{
    // close socket
    _mcastSocket.close();
    // close socket
    _acceptSocket.close();
}

/*! get connection type
 */
const ConnectionType *PointMCastConnection::getType() const
{
    return &_type;
}

/*-------------------------------------------------------------------------*/
/*                            connection                                   */

/*! connect to the given group. If timeout is reached, -1 is
    returned
 */
Connection::Channel PointMCastConnection::connectGroup(
    const std::string &address,
    Time               timeout)
{
    Channel channel = Inherited::connectGroup(address,timeout);
    return channel;
}

/*! disconnect the given channel
 */
void PointMCastConnection::disconnect(void)
{
    _socket.close();
}

/*! accept an icomming grop connection. If timeout is reached,
    -1 is returned. If timeout is -1 then wait without timeout
 */
Connection::Channel PointMCastConnection::acceptGroup(Time timeout)
{
    Channel channel = Inherited::acceptGroup(timeout);
    return channel;
}

/*-------------------------------------------------------------------------*/
/*                              channel handling                           */

/*! select the next channel for reading. If timeout is not -1
    then -1 is returned if timeout is reached
*/
Connection::Channel PointMCastConnection::selectChannel(Time timeout)
    throw (ReadError)
{
    if(_pointToPoint)
        return Inherited::selectChannel(timeout);
    // TODO FIXME
    //_mcastSocket.waitReadable(.1);
    return 0;
}

int PointMCastConnection::getAvailable(void)
{
    if(_pointToPoint)
        return Inherited::getAvailable();
    return _mcastSocket.getAvailable();
}

/*-------------------------------------------------------------------------*/
/*                            sync                                         */

/*! wait for signal
 */
bool PointMCastConnection::wait(Time timeout) throw (ReadError)
{
    UInt32 tag;

    if(_pointToPoint)
        return Inherited::wait(timeout);
    try
    {
        if(selectChannel(timeout) < 0)
            return false;
        getValue(tag);
        if(tag != 314156)
        {
            FFATAL(("Stream out of sync in PointMCastConnection\n"));
            throw ReadError("Stream out of sync");
        }
    }
    catch(SocketError &e)
    {
        throw ReadError(e.what());
    }
    return true;
}

/*-------------------------- create ---------------------------------------*/

/** \brief create conneciton
 */

PointConnection *PointMCastConnection::create(void)
{
    return new PointMCastConnection();
}

/*-------------------------------------------------------------------------*/
/*                              read write                                 */

/** Read data into given memory
 *
 * Read data form the current read socket. The read socket is that
 * socket, that was selectet in selectChannel.
 *
 **/

SizeT PointMCastConnection::read(MemoryHandle mem, SizeT size)
{
    if(_pointToPoint)
    {
        return Inherited::read(mem,size);
    }

    if(!_initialized)
        initialize();

    // without this we get a deadlock on cluster stop!
    if(_socket.isClosed())
    {
        throw ReadError("Socket was closed!");
    }
    UInt32 count_id = 0;
    _socket.recv(&count_id, sizeof(count_id));

    UInt16 write_id = (count_id >> 16);
    UInt32 count = (count_id & 0xffff);

    //UInt32 ack = 0;
    //_socket.send(&ack, sizeof(ack));

    //printf("try receive %u dgrams\n", count);
    //fflush(stdout);

    bool compression = false;
    SizeT       rsize   = 0;
    MemoryHandle buffer = mem;
    if(count > _interactive_threshold)
    {
        SocketAddress from;
        // we need to use a map here to detect dupicated dgrams!
        std::map<UInt32, Dgram *> dgrams;
        //Dgram *dgram = new Dgram;
        Dgram *dgram = _dgramPool.alloc();

        UInt32 iterations = count;
        while(iterations)
        {
            //if(_mcastSocket.getAvailable() > 0)
            if(_mcastSocket.waitReadable(0.1))
            {
                if(dgram == NULL)
                {
                    //dgram = new Dgram;
                    dgram = _dgramPool.alloc();
                }

                UInt32 length = _mcastSocket.recvFrom(dgram->getBuffer(),
                                                      dgram->getBufferCapacity(),
                                                      from);
                dgram->setBufferSize(length);

                //printf("received dgram %d size %d\n", dgram->getId(), dgram->getSize());
                //fflush(stdout);

                // look for a valid dgram.
                // could be a "lost" dgram from the last readBuffer() so we filter out wrong id's!
                // filter out dgrams from another sender from == _sender
                if(from == _sender && dgram->getBaseId() == write_id && dgram->getId() < count && dgrams.count(dgram->getId()) == 0)
                {
                    if(dgram->getFlags() & Dgram::DGRAM_LAST)
                        compression = (dgram->getFlags() & Dgram::DGRAM_COMPRESSED);
                    dgrams.insert(std::make_pair(dgram->getId(), dgram));
                    dgram = NULL;
                    --iterations;
                }
            }
            else
            {
                break;
            }
        }

        if(dgram != NULL)
            _dgramPool.free(dgram);

        UInt32 ack_count = dgrams.size();
        _socket.send(&ack_count, sizeof(ack_count));

        //printf("got %u/%u dgrams\n", ((UInt32) dgrams.size()), count);
        //fflush(stdout);

        //UInt32       size   = readBufBegin()->getSize();
        //printf("read buffer size %d\n", readBufBegin()->getSize());

        for(UInt32 id=0;id<count;++id)
        {
            std::map<UInt32, Dgram *>::iterator it = dgrams.find(id);
            if(it == dgrams.end())
                dgram = NULL;
            else
                dgram = it->second;

            if(dgram == NULL)
            {
                // receive missing dgram via unicast.
                //printf("ask for missing dgram %u\n", id);
                //fflush(stdout);
                _socket.send(&id, sizeof(id));

                UInt32 dgram_size = 0;
                _socket.recv(&dgram_size, sizeof(dgram_size));
                //dgram = new Dgram;
                dgram = _dgramPool.alloc();
                _socket.recv(dgram->getBuffer(), dgram_size);
                dgram->setBufferSize(dgram_size);

                if(dgram->getFlags() & Dgram::DGRAM_LAST)
                    compression = (dgram->getFlags() & Dgram::DGRAM_COMPRESSED);

                //printf("got missing dgram %u size %u\n", dgram->getId(), dgram->getSize());
                //fflush(stdout);
                //UInt32 ack = 0;
                //_socket.send(&ack, sizeof(ack));
            }
            memcpy(buffer, dgram->getData(), dgram->getSize());
            rsize += dgram->getSize();
            buffer += dgram->getSize();
            //delete dgram;
            _dgramPool.free(dgram);
        }
    }
    else
    {
        Dgram *dgram = _dgramPool.alloc();
        for(UInt32 id=0;id<count;++id)
        {
            UInt32 dgram_size = 0;
            _socket.recv(&dgram_size, sizeof(dgram_size));
            _socket.recv(dgram->getBuffer(), dgram_size);
            dgram->setBufferSize(dgram_size);
            if(dgram->getFlags() & Dgram::DGRAM_LAST)
                compression = (dgram->getFlags() & Dgram::DGRAM_COMPRESSED);
            memcpy(buffer, dgram->getData(), dgram->getSize());
            rsize += dgram->getSize();
            buffer += dgram->getSize();
        }
        _dgramPool.free(dgram);
    }

    return rsize;
}

void PointMCastConnection::readBuffer()
{
    if(_pointToPoint)
    {
        Inherited::readBuffer();
        return;
    }

    if(!_initialized)
        initialize();

    // without this we get a deadlock on cluster stop!
    if(_socket.isClosed())
    {
        throw ReadError("Socket was closed!");
    }
    UInt32 count_id = 0;
    _socket.recv(&count_id, sizeof(count_id));

    UInt16 write_id = (count_id >> 16);
    UInt32 count = (count_id & 0xffff);

    bool compression = false;
    UInt32       size   = 0;
    MemoryHandle buffer = readBufBegin()->getMem();

    if(count > _interactive_threshold)
    {
        //UInt32 ack = 0;
        //_socket.send(&ack, sizeof(ack));

        //printf("try receive %u dgrams\n", count);
        //fflush(stdout);

        SocketAddress from;
        // we need to use a map here to detect duplicated dgrams!
        std::map<UInt32, Dgram *> dgrams;
        //Dgram *dgram = new Dgram;
        Dgram *dgram = _dgramPool.alloc();

        UInt32 iterations = count;
        while(iterations)
        {
            //if(_mcastSocket.getAvailable() > 0)
            if(_mcastSocket.waitReadable(0.1))
            {
                if(dgram == NULL)
                {
                    //dgram = new Dgram;
                    dgram = _dgramPool.alloc();
                }

                UInt32 length = _mcastSocket.recvFrom(dgram->getBuffer(),
                                                      dgram->getBufferCapacity(),
                                                      from);
                dgram->setBufferSize(length);

                //printf("received dgram %d size %d\n", dgram->getId(), dgram->getSize());
                //fflush(stdout);

                // look for a valid dgram.
                // could be a "lost" dgram from the last readBuffer() so we filter out wrong id's!
                if(from == _sender && dgram->getBaseId() == write_id && dgram->getId() < count && dgrams.count(dgram->getId()) == 0)
                {
                    if(dgram->getFlags() & Dgram::DGRAM_LAST)
                        compression = (dgram->getFlags() & Dgram::DGRAM_COMPRESSED);
                    dgrams.insert(std::make_pair(dgram->getId(), dgram));
                    dgram = NULL;
                    --iterations;
                }
            }
            else
            {
                break;
            }
        }

        if(dgram != NULL)
            _dgramPool.free(dgram);

        UInt32 ack_count = dgrams.size();
        _socket.send(&ack_count, sizeof(ack_count));

        //printf("got %u/%u dgrams\n", ((UInt32) dgrams.size()), count);
        //fflush(stdout);

        //UInt32       size   = readBufBegin()->getSize();
        //printf("read buffer size %d\n", readBufBegin()->getSize());

        for(UInt32 id=0;id<count;++id)
        {
            std::map<UInt32, Dgram *>::iterator it = dgrams.find(id);
            if(it == dgrams.end())
                dgram = NULL;
            else
                dgram = it->second;

            if(dgram == NULL)
            {
                // receive missing dgram via unicast.
                //printf("ask for missing dgram %u\n", id);
                //fflush(stdout);
                _socket.send(&id, sizeof(id));

                UInt32 dgram_size = 0;
                _socket.recv(&dgram_size, sizeof(dgram_size));
                //dgram = new Dgram;
                dgram = _dgramPool.alloc();
                _socket.recv(dgram->getBuffer(), dgram_size);
                dgram->setBufferSize(dgram_size);

                if(dgram->getFlags() & Dgram::DGRAM_LAST)
                    compression = (dgram->getFlags() & Dgram::DGRAM_COMPRESSED);

                //printf("got missing dgram %u size %u\n", dgram->getId(), dgram->getSize());
                //fflush(stdout);
                //UInt32 ack = 0;
                //_socket.send(&ack, sizeof(ack));
            }
            memcpy(buffer, dgram->getData(), dgram->getSize());
            size += dgram->getSize();
            buffer += dgram->getSize();
            //delete dgram;
            _dgramPool.free(dgram);
        }
    }
    else
    {
        Dgram *dgram = _dgramPool.alloc();
        for(UInt32 id=0;id<count;++id)
        {
            UInt32 dgram_size = 0;
            _socket.recv(&dgram_size, sizeof(dgram_size));
            _socket.recv(dgram->getBuffer(), dgram_size);
            dgram->setBufferSize(dgram_size);
            if(dgram->getFlags() & Dgram::DGRAM_LAST)
                compression = (dgram->getFlags() & Dgram::DGRAM_COMPRESSED);
            memcpy(buffer, dgram->getData(), dgram->getSize());
            size += dgram->getSize();
            buffer += dgram->getSize();
        }
        _dgramPool.free(dgram);
    }

    //printf("size %d compression %d\n", size, compression);
    //fflush(stdout);

    readBufBegin()->setDataSize(size);

    /*
    CRC32 crc;
    UInt32 checksum = crc.fullCRC(readBufBegin()->getMem(), readBufBegin()->getDataSize());
    UInt32 mchecksum = 0;
    _socket.recv(&mchecksum, sizeof(mchecksum));
    if(mchecksum != checksum)
    {
        printf("wrong checksum %x != %x\n", mchecksum, checksum);
        fflush(stdout);
    }
    */

    // now look for a complete buffer!
    if(_compression == Connection::CCOMPRESSION_BEST && compression)
    {
        static std::vector<UInt8> mem;
        mem.resize(readBufBegin()->getDataSize());
        memcpy(&mem[0], readBufBegin()->getMem(), mem.size());
        mz_ulong ucsize = _ioBufferSize;
        int r = mz_uncompress((unsigned char *) readBufBegin()->getMem(), &ucsize, (const unsigned char *) &mem[0], mem.size());
        if(r != 0)
            printf("PointMcastConnection::readBuffer: uncompress error!\n");
        readBufBegin()->setDataSize(ucsize);
    }
    // now look for a complete buffer!
    else if(_compression == Connection::CCOMPRESSION_FAST && compression)
    {
        static std::vector<UInt8> mem;
        mem.resize(readBufBegin()->getDataSize());
        memcpy(&mem[0], readBufBegin()->getMem(), mem.size());
        int ucsize = LZ4_decompress_safe((const char *) &mem[0], (char *) readBufBegin()->getMem(), mem.size(), readBufBegin()->getSize());
        readBufBegin()->setDataSize(ucsize);
    }
}

/*-------------------------------------------------------------------------*/
/*                              private helpers                            */

/*! initialize connection
 */
void PointMCastConnection::initialize()
{
    std::string   group;
    Channel       channel;
    BinaryMessage message;
    std::string   fromHost;
    UInt32        fromPort;
    UInt32        combineCount;
    std::string   host;
    UInt32        port;
    char          threadName[256];

    sprintf(threadName,"PointMCastConnection%p",this);

    // get info about the group
    _socket.recv(message);
    // group and port
    group = message.getString();
    port  = message.getUInt32();
    // get port
    fromPort = message.getUInt32();

    _sender = SocketAddress(_remoteAddress.getHost().c_str(), fromPort);

    std::cout << _remoteAddress.getHost() << " " << fromPort << std::endl;

    // prepare socket to receive mcast packages
    _mcastSocket.open();
    _mcastSocket.setReusePort(true);
    _mcastSocket.setReadBufferSize(524288);
    _mcastSocket.bind(SocketAddress(SocketAddress::ANY, port));

    try
    {
        _mcastSocket.join(SocketAddress(group.c_str()));
    }
    catch(...)
    {
    }

    // set multicast interface
    if(!getInterface().empty())
    {
        _mcastSocket.setMCastInterface(SocketAddress(getInterface().c_str()));
    }

    _initialized = true;
    //std::cout << "init end" << std::endl;
}

/*-------------------------------------------------------------------------*/
/*                              static type                                */

ConnectionType PointMCastConnection::_type(
    &PointMCastConnection::create,
    "Multicast");

/*-------------------------------------------------------------------------*/
/*                              cvs id's                                   */

#ifdef __sgi
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning( disable : 177 )
#endif

namespace
{
    static Char8 cvsid_cpp       [] = "@(#)$Id: $";
    static Char8 cvsid_hpp       [] = OSG_GROUPMCASTCONNECTION_HEADER_CVSID;
}

