/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *             Copyright (C) 2000-2002 by the OpenSG Forum                   *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

//---------------------------------------------------------------------------
//  Includes
//---------------------------------------------------------------------------
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "OSGConfig.h"
#include "OSGLog.h"
#include "OSGBaseFunctions.h"
#include "OSGSocketSelection.h"
#include "OSGPointSockConnection.h"
#include "OSGConnectionHelper.h"
#include "OSGConnectionType.h"

#include "OSGminiz.h"
#include "OSGLZ4.h"

#pragma intrinsic(memcpy)

OSG_USING_NAMESPACE

/** \class osg::PointSockConnection
 *  \ingroup GrpBaseNetwork
 **/

// amz changed from 131071 to 131072
/* static */ const UInt32 PointSockConnection::_ioBufferSize = 131072;

/*-------------------------------------------------------------------------*/
/*                            constructor destructor                       */

/*! Constructor
 */

PointSockConnection::PointSockConnection(void) :
    Inherited         (0),
    _acceptSocket     (),
    _socket           (),
    _remoteAddress    (),
    _socketReadBuffer (),
    _socketWriteBuffer()
{
    _acceptSocket.open();
    _acceptSocket.setReusePort(true);

    _socketReadBuffer .resize(_ioBufferSize);
    _socketWriteBuffer.resize(_ioBufferSize);

    // reserve first bytes for buffer size
    readBufAdd (&_socketReadBuffer [sizeof(SocketBufferHeader)],
                _socketReadBuffer.size() - sizeof(SocketBufferHeader));
    writeBufAdd(&_socketWriteBuffer[sizeof(SocketBufferHeader)],
                _socketWriteBuffer.size() - sizeof(SocketBufferHeader));
}

/*! Destructor
 */
PointSockConnection::~PointSockConnection(void)
{
    _acceptSocket.close();

    if(_socket.isOpen())
        _socket.close();
}

/*! get connection type
 */
const ConnectionType *PointSockConnection::getType(void) const
{
    return &_type;
}

/*-------------------------------------------------------------------------*/
/*                            connection                                   */

/*! Connect to the given point. If timeout is reached, -1 is
    returned
 */
Connection::Channel PointSockConnection::connectPoint(
    const std::string &address,
    Time               timeout)
{
    StreamSocket socket;
    if(connectSocket(socket,         address,
                     _remoteAddress, timeout ))
    {
        _socket = socket;
        _pointToPoint = true;
        return 0;
    }
    else
    {
        return -1;
    }
}

/*! Connect to the given group. If timeout is reached, -1 is
    returned
 */
Connection::Channel PointSockConnection::connectGroup(
    const std::string &address,
    Time               timeout)
{
    StreamSocket socket;
    if(connectSocket(socket,         address,
                     _remoteAddress, timeout ))
    {
        _socket = socket;
        _pointToPoint = false;
        return 0;
    }
    else
    {
        return -1;
    }
}

/*! Accept an icomming point connection. If timeout is reached,
    -1 is returned. If timeout is -1 then wait without timeout
 */
Connection::Channel PointSockConnection::acceptPoint(Time timeout)
{
    if(acceptSocket(_acceptSocket,  _socket,
                    _remoteAddress, timeout ))
    {
        _pointToPoint = true;
        setLocalInterface(_socket.getAddress().getHost());
        setRemoteInterface(_remoteAddress.getHost());
        return 0;
    }
    else
    {
        return -1;
    }
}

/*! Accept an icomming grop connection. If timeout is reached,
    -1 is returned. If timeout is -1 then wait without timeout
 */
Connection::Channel PointSockConnection::acceptGroup(Time timeout)
{
    if(acceptSocket(_acceptSocket,  _socket,
                    _remoteAddress, timeout ))
    {
        _pointToPoint = false;
        setLocalInterface(_socket.getAddress().getHost());
        setRemoteInterface(_remoteAddress.getHost());
        return 0;
    }
    else
    {
        setLocalInterface("");
        setRemoteInterface("");
        return -1;
    }
}

/*! Disconnect the given channel
 */
void PointSockConnection::disconnect(void)
{
    _socket.close();
}
/*! bind the connection to a network interface. The address is
    returned, on wich the port could be connected. The interface
    is determined by the connection interface filed and the
    address parameter. Address can be empty, wich means to use
    a free port or address can contain a port number.
 */
std::string PointSockConnection::bind(const std::string &address)
{
    int         port = 0;
    char        localhost[256];
    char        host[256];
    char        portStr[256];
    std::string interf = address;

    // remove port from interface!
    std::string::size_type p = interf.find(":");
    if(p != std::string::npos)
        interf = interf.substr(0, p);

    if(!getInterface().empty())
    {
        interf = getInterface();
    }

    if(interf.empty())
    {
        // get local host name
        osgGetHostname(localhost, 255);
        interf = localhost;
    }

    // parse address
    if(!address.empty())
        if(sscanf(address.c_str(), "%*[^:]:%d", &port) != 1)
            if(sscanf(address.c_str(), ":%d", &port) != 1)
                port = 0;
    // bind port
    _acceptSocket.setReusePort(true);

    if(!interf.empty())
        _acceptSocket.bind(SocketAddress(interf.c_str(), port));
    else
        _acceptSocket.bind(SocketAddress(SocketAddress::ANY, port));

    SINFO << "Connection bound to "
          << _acceptSocket.getAddress().getHost() << ":"
          << _acceptSocket.getAddress().getPort() << std::endl;
    _acceptSocket.listen();
    // create address
    sprintf(portStr, "%d", _acceptSocket.getAddress().getPort());
    return _acceptSocket.getAddress().getHost() + ":" + portStr;
}

/*-------------------------------------------------------------------------*/
/*                              channel handling                           */

/*! select the next channel for reading. If timeout is not -1
    then -1 is returned if timeout is reached
*/
Connection::Channel PointSockConnection::selectChannel(Time timeout)
throw (ReadError)
{
    try
    {
        if(_socket.waitReadable(timeout))
            return 0;
    }
    catch(SocketError &e)
    {
        throw ReadError(e.what());
    }
    return -1;
}

int PointSockConnection::getAvailable(void)
{
    return _socket.getAvailable();
}

/*-------------------------------------------------------------------------*/
/*                            sync                                         */

/*! wait for signal
 */
bool PointSockConnection::wait(Time timeout) throw (ReadError)
{
    UInt32 tag;
    try
    {
        if(!_socket.waitReadable(timeout))
            return false;
        if(!_socket.recv(&tag, sizeof(tag)))
            throw ReadError("Channel closed");
        tag = osgntohl(tag);
        if(tag != 314156)
        {
            FFATAL(("Stream out of sync in SockConnection (tag is %u)\n", tag));
            throw ReadError("Stream out of sync");
        }
    }
    catch(SocketError &e)
    {
        throw ReadError(e.what());
    }
    return true;
}

/*! send signal
 */
void PointSockConnection::signal(void) throw (WriteError)
{
    UInt32 tag = osghtonl(314156);
    try
    {
        _socket.send(&tag, sizeof(tag));
    }
    catch(SocketError &e)
    {
        throw WriteError(e.what());
    }
}

/*-------------------------- create ---------------------------------------*/

/** \brief create conneciton
 */

PointConnection *PointSockConnection::create(void)
{
    return new PointSockConnection();
}

/*-------------------------------------------------------------------------*/
/*                              read write                                 */

/** Read data into given memory
 *
 * Read data form the current read socket. The read socket is that
 * socket, that was selectet in selectChannel.
 *
 **/

SizeT PointSockConnection::read(MemoryHandle mem, SizeT size)
{
    int len = 0;

    // read data
    len = _socket.recv(mem, size);
    if(len == 0)
    {
        throw ReadError("read got 0 bytes!");
    }

    return len;
}

/** Read next data block
 *
 * The stream connection uses only BinaryDataHandler buffer. If more
 * then one buffer is present, then this methode must be changed!
 *
 */

void PointSockConnection::readBuffer()
{
    int size;
    int len;

    // read buffer header
    len = _socket.recv(&_socketReadBuffer[0], sizeof(SocketBufferHeader));
    if(len == 0)
        throw ReadError("peek got 0 bytes!");

    // read remaining data
    size = osgntohl(reinterpret_cast<SocketBufferHeader *>(&_socketReadBuffer[0])->size);
    UInt32 rsize = size;

    UInt32 compressed_size = osgntohl(reinterpret_cast<SocketBufferHeader *>(&_socketReadBuffer[0])->compressed_size);
    if(compressed_size > 0)
        rsize = compressed_size;

    len = _socket.recv(&_socketReadBuffer[sizeof(SocketBufferHeader)], rsize);
    if(len == 0)
        throw ReadError("read got 0 bytes!");

    if(_compression == Connection::CCOMPRESSION_BEST && compressed_size > 0)
    {
        static std::vector<UInt8> mem;
        mem.resize(rsize);
        memcpy(&mem[0], &_socketReadBuffer[sizeof(SocketBufferHeader)], mem.size());
        mz_ulong ucsize = _ioBufferSize;
        int r = mz_uncompress((unsigned char *) &_socketReadBuffer[sizeof(SocketBufferHeader)], &ucsize, (const unsigned char *) &mem[0], mem.size());
        if(r != 0)
            printf("PointSockConnection::readBuffer: uncompress error!\n");
        size = ucsize;
    }
    else if(_compression == Connection::CCOMPRESSION_FAST && compressed_size > 0)
    {
        static std::vector<UInt8> mem;
        mem.resize(rsize);
        memcpy(&mem[0], &_socketReadBuffer[sizeof(SocketBufferHeader)], rsize);
        int ucsize = LZ4_decompress_safe((const char *) &mem[0], (char *) &_socketReadBuffer[sizeof(SocketBufferHeader)], rsize, size);
        size = ucsize;
        //printf("PointSockConnection::readBuffer: uncompressed %d ==> %d\n", len, size);
    }
    else
    {
        //printf("PointSockConnection::readBuffer: uncompressed %d\n", len);
    }

    readBufBegin()->setDataSize(size);
}

/** Write data to destinations
 *
 * \param mem   Pointer to data buffer
 * \param size  Size of bytes to write
 *
 **/

void PointSockConnection::write(const MemoryHandle mem, SizeT size)
{
    _socket.send(mem, size);
}

/** Write buffer
 *
 * Write blocksize and data.
 *
 **/
void PointSockConnection::writeBuffer(void)
{
    Int32 index;
    UInt32 size = writeBufBegin()->getDataSize();
    // write size to header
    reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0])->size = osghtonl(size);
    reinterpret_cast<SocketBufferHeader *>(&_socketWriteBuffer[0])->compressed_size = osghtonl(0);
    UInt8 *data = &_socketWriteBuffer[0];

    if(size)
    {
        if(_compression == Connection::CCOMPRESSION_BEST && size > 1)
        {
            static std::vector<UInt8> compressed;
            compressed.resize(mz_compressBound(size) + sizeof(SocketBufferHeader));
            mz_ulong compressed_size = compressed.size() - sizeof(SocketBufferHeader);
            int r = mz_compress2((unsigned char *) &compressed[sizeof(SocketBufferHeader)], &compressed_size, (const unsigned char *) &_socketWriteBuffer[sizeof(SocketBufferHeader)], size, 1);
            if(r == 0 && compressed_size != 0 && compressed_size < size)
            {
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->size = osghtonl(size);
                size = compressed_size;
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->compressed_size = osghtonl(size);
                data = &compressed[0];
            }
            else
            {
                //printf("GroupSockConnection::writeBuffer: send uncompressed %d!\n", size);
            }
        }
        else if(_compression == Connection::CCOMPRESSION_FAST && size > 1)
        {
            static std::vector<UInt8> compressed;
            // ok we allocate size - 1, lzf_compress returns 0 when the compression failed,
            // in this case the compressed data is greater than the input data.
            compressed.resize(LZ4_compressBound(size) + sizeof(SocketBufferHeader));
            int compressed_size = LZ4_compress((const char *) &_socketWriteBuffer[sizeof(SocketBufferHeader)], (char *) &compressed[sizeof(SocketBufferHeader)], size);

            if(compressed_size != 0 && compressed_size < size)
            {
                //printf("PointSockConnection::writeBuffer: compressed %d ==> %d\n", size, compressed_size);
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->size = osghtonl(size);
                size = compressed_size;
                reinterpret_cast<SocketBufferHeader *>(&compressed[0])->compressed_size = osghtonl(size);
                data = &compressed[0];
            }
            else
            {
                //printf("GroupSockConnection::writeBuffer: send uncompressed %d!\n", size);
            }
        }
        else
        {
            //printf("GroupSockConnection::writeBuffer: send uncompressed %d!\n", size);
        }

        // write whole block
        _socket.send(data, size + sizeof(SocketBufferHeader));
    }
}

void PointSockConnection::setCork(bool s)
{
    _socket.setCork(s);
}

void PointSockConnection::setDelay(bool s)
{
    _socket.setDelay(s);
}

int PointSockConnection::waitAvailable(void)
{
    return _socket.waitAvailable();
}

/*-------------------------------------------------------------------------*/
/*                              static type                                */

ConnectionType PointSockConnection::_type(
    &PointSockConnection::create,
    "StreamSock");

/*-------------------------------------------------------------------------*/
/*                              cvs id's                                   */

#ifdef __sgi
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning( disable : 177 )
#endif

namespace
{
    static Char8 cvsid_cpp       [] = "@(#)$Id: $";
    static Char8 cvsid_hpp       [] = OSG_GROUPSOCKCONNECTION_HEADER_CVSID;
}

