/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *             Copyright (C) 2000-2002 by the OpenSG Forum                   *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

//#define MULTI_THREADED
#define THRESHOLD_SUPPORT

//---------------------------------------------------------------------------
//  Includes
//---------------------------------------------------------------------------
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "OSGConfig.h"
#include "OSGLog.h"
#include "OSGBaseFunctions.h"
#include "OSGBaseThread.h"
#include "OSGSocketSelection.h"
#include "OSGBinaryMessage.h"
#include "OSGPointSockPipeline.h"
#include "OSGGroupSockPipeline.h"
#include "OSGConnectionType.h"
#include "OSGBinaryMessage.h"

#include "OSGminiz.h"
#include "OSGLZ4.h"

OSG_USING_NAMESPACE

/** \class osg::PointSockPipeline
 *  \ingroup GrpBaseNetwork
 **/

/*-------------------------------------------------------------------------*/
/*                            constructor destructor                       */

/*! Constructor
 */

PointSockPipeline::PointSockPipeline():
    Inherited(),
    _initialized(false),
    _next(),
    _prev(),
    _last(false),
    _tp()
{
    _prev.open();
    _next.open();
}

/*! Destructor
 */
PointSockPipeline::~PointSockPipeline(void)
{
    _prev.close();
    _next.close();
}

/*! get connection type
 */
const ConnectionType *PointSockPipeline::getType() const
{
    return &_type;
}

/*-------------------------------------------------------------------------*/
/*                            connection                                   */

/*! connect to the given group. If timeout is reached, -1 is
    returned
 */
Connection::Channel PointSockPipeline::connectGroup(
    const std::string &address,
    Time               timeout)
{
    Channel channel = Inherited::connectGroup(address, timeout);
    return channel;
}

/*! disconnect the given channel
 */
void PointSockPipeline::disconnect(void)
{
    _socket.close();
}

/*! accept an icomming grop connection. If timeout is reached,
    -1 is returned. If timeout is -1 then wait without timeout
 */
Connection::Channel PointSockPipeline::acceptGroup(Time timeout)
{
    Channel channel = Inherited::acceptGroup(timeout);
    return channel;
}

/*-------------------------------------------------------------------------*/
/*                              channel handling                           */

/*! select the next channel for reading. If timeout is not -1
    then -1 is returned if timeout is reached
*/
Connection::Channel PointSockPipeline::selectChannel(Time timeout)
throw (ReadError)
{
    if(_pointToPoint)
        return Inherited::selectChannel(timeout);

    if(!_initialized)
        initialize();

    try
    {
#ifdef THRESHOLD_SUPPORT
        timeout = 0.0;
        while(true)
        {
            bool s1 = _socket.waitReadable(timeout);
            bool s2 = _prev.waitReadable(timeout);
            if(s1 || s2)
                return 0;
        }
#else
        if(_prev.waitReadable(timeout))
            return 0;
#endif
    }
    catch(SocketError &e)
    {
        throw ReadError(e.what());
    }
    return -1;
}

/*-------------------------- create ---------------------------------------*/

/** \brief create conneciton
 */

PointConnection *PointSockPipeline::create(void)
{
    return new PointSockPipeline();
}

/*-------------------------------------------------------------------------*/
/*                              read write                                 */

/** Read data into given memory
 *
 * Read data form the current read socket. The read socket is that
 * socket, that was selectet in selectChannel.
 *
 **/

SizeT PointSockPipeline::read(MemoryHandle mem, SizeT size)
{
    if(_pointToPoint)
    {
        return Inherited::read(mem, size);
    }

    int len = 0;

    if(!_initialized)
        initialize();
    // read data
    len = _prev.recv(mem, size);
    if(len == 0)
    {
        throw ReadError("read got 0 bytes!");
    }
    // send to next in chain
    if(!_last)
        _next.send(mem, size);

    return len;
}

typedef struct
{
    std::vector<UInt8> buffer;
    PointSockPipeline *the;
} ThreadData;

/** Read next data block
 *
 * The stream connection uses only BinaryDataHandler buffer. If more
 * then one buffer is present, then this methode must be changed!
 *
 */

void PointSockPipeline::readBuffer()
{
    if(_pointToPoint)
    {
        Inherited::readBuffer();
        return;
    }

#ifdef THRESHOLD_SUPPORT

    static UInt32 last_id = 0;

    if(_prev.getAvailable() == 0 && _socket.getAvailable() == 0)
    {
        while(true)
        {
            //bool s1 = _socket.waitReadable(0.0);
            //bool s2 = _prev.waitReadable(0.0);
            bool s1 = (_socket.getAvailable() > 0);
            bool s2 = (_prev.getAvailable() > 0);
            if(s1 || s2)
                break;
            osgyield();
        }
    }

    // look for the end marker
    if(_socket.getAvailable() == sizeof(UInt32))
    {
        UInt32 end;
        _socket.peek(&end, sizeof(end));
        if(end == 0x87654321)
        {
            throw ReadError("detected end!");
        }
    }

    // via id we look for the right order could be a unicast (threshold) or
    // pipeline package.
    bool found_prev_id = false;
    if(_prev.getAvailable() > 0)
    {
        UInt32 header[2];
        _prev.peek(&header, sizeof(header));
        if(header[1] == last_id + 1)
            found_prev_id = true;
    }

    if(!found_prev_id && _socket.getAvailable() > 0)
    {
        UInt32 header[2];
        _socket.peek(&header, sizeof(header));
        if(header[0] == 0x12345678)
        {
            if(header[1] == last_id + 1)
            {
                Inherited::readBuffer();
                last_id = header[1];
                return;
            }
        }
    }

#endif

    int size;
    int len;

    if(!_initialized)
        initialize();

    // read buffer header
    len = _prev.recv(&_socketReadBuffer[0], sizeof(SocketBufferHeader));
    if(len == 0)
        throw ReadError("peek got 0 bytes!");

#ifdef THRESHOLD_SUPPORT
    UInt32 id = reinterpret_cast<SocketBufferHeader *>(&_socketReadBuffer[0])->id;
    if(id != last_id + 1)
    {
        throw ReadError("PointSockPipeline::readBuffer: found wrong PIPELINE packet id!");
    }
    last_id = id;
#endif

    // read remaining data
    size = osgntohl(reinterpret_cast<SocketBufferHeader *>(&_socketReadBuffer[0])->size);
    UInt32 rsize = size;

    UInt32 compressed_size = osgntohl(reinterpret_cast<SocketBufferHeader *>(&_socketReadBuffer[0])->compressed_size);
    if(compressed_size > 0)
        rsize = compressed_size;

    len = _prev.recv(&_socketReadBuffer[sizeof(SocketBufferHeader)], rsize);
    if(len == 0)
        throw ReadError("read got 0 bytes!");

    // send to next in chain
    if(!_last)
    {
#ifdef MULTI_THREADED
        // send the compressed stream to the next machine in a extra thread
        // so we can uncompress it in parallel for the current machine.
        _tp.wait();

        static ThreadData *td = new ThreadData;
        td->buffer.resize(rsize + sizeof(SocketBufferHeader));
        memcpy(&td->buffer[0], &_socketReadBuffer[0], rsize + sizeof(SocketBufferHeader));
        td->the = this;
        _tp.run(writeThread, static_cast<void *>(td));
#else
        _next.send(&_socketReadBuffer[0], sizeof(SocketBufferHeader) + rsize);
#endif
    }

    if(_compression == Connection::CCOMPRESSION_BEST && compressed_size > 0)
    {
        static std::vector<UInt8> mem;
        mem.resize(rsize);
        memcpy(&mem[0], &_socketReadBuffer[sizeof(SocketBufferHeader)], mem.size());
        mz_ulong ucsize = _ioBufferSize;
        int r = mz_uncompress((unsigned char *) &_socketReadBuffer[sizeof(SocketBufferHeader)], &ucsize, (const unsigned char *) &mem[0], mem.size());
        if(r != 0)
            printf("PointSockConnection::readBuffer: uncompress error!\n");
        size = ucsize;
    }
    else if(_compression == Connection::CCOMPRESSION_FAST && compressed_size > 0)
    {
        static std::vector<UInt8> mem;
        mem.resize(rsize);
        memcpy(&mem[0], &_socketReadBuffer[sizeof(SocketBufferHeader)], rsize);
        int ucsize = LZ4_decompress_safe((const char *) &mem[0], (char *) &_socketReadBuffer[sizeof(SocketBufferHeader)], rsize, size);
        size = ucsize;
        //printf("PointSockConnection::readBuffer: uncompressed %d ==> %d\n", len, size);
    }
    else
    {
        //printf("PointSockConnection::readBuffer: uncompressed %d\n", len);
    }

    readBufBegin()->setDataSize(size);
}

/*! write thread
 */
void PointSockPipeline::writeThread(void *arg)
{
    ThreadData *td = (ThreadData *) arg;
    td->the->_next.send(&td->buffer[0], td->buffer.size());
    // we use one thread only so one td instance is enough!
    //delete td;
}

/*-------------------------------------------------------------------------*/
/*                              private helpers                            */

/*! initialize pipeline
 */
void PointSockPipeline::initialize(void)
{
    BinaryMessage message;
    StreamSocket  sock;
    UInt32        nextPort;
    std::string   nextHost;
    UInt32        len;
    std::string   interf;

    // get local host name doesn't work in all cases we need a ip
    // so we already have a open connection via _socket and we read the ip from there!
    //osgGetHostname(localhost, 255);
    if(!getInterface().empty())
        interf = getInterface();
    else
        interf = _socket.getAddress().getHost();

    sock.open();
    sock.bind(SocketAddress(interf.c_str(), 0));
    sock.listen();

    // send my own address
    message.putString(interf);
    message.putUInt32(sock.getAddress().getPort());
    _socket.send(message);
    // accept prev
    _prev = sock.accept();
    sock.close();

    len = _socket.recv(message);
    if(len == 0)
        throw ReadError("Channel closed\n");
    _last = message.getUInt32();
    if(!_last)
    {
        nextHost = message.getString();
        nextPort = message.getUInt32();
        for(;;)
        {
            try
            {
                _next.connect(SocketAddress(nextHost.c_str(),
                                            nextPort));
                break;
            }
            catch(...)
            {
            }
        }
    }

    _initialized = true;
}

/*-------------------------------------------------------------------------*/
/*                              static type                                */

ConnectionType PointSockPipeline::_type(
    &PointSockPipeline::create,
    "SockPipeline");

/*-------------------------------------------------------------------------*/
/*                              cvs id's                                   */

#ifdef __sgi
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning( disable : 177 )
#endif

namespace
{
    static Char8 cvsid_cpp       [] = "@(#)$Id: $";
    static Char8 cvsid_hpp       [] = OSG_GROUPSOCKPIPELINE_HEADER_CVSID;
}

